"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Task = void 0;
const tslib_1 = require("tslib");
const child_process_1 = require("child_process");
const fs_1 = require("fs");
const path_1 = require("path");
const util_1 = require("util");
const chalk_1 = tslib_1.__importDefault(require("chalk"));
const logging_1 = require("./logging");
const ENV_TRIM_LEN = 20;
class Task {
    constructor(runtime, task, parents = []) {
        var _a, _b;
        this.runtime = runtime;
        this.task = task;
        this.env = {};
        this.workdir = (_a = task.cwd) !== null && _a !== void 0 ? _a : this.runtime.workdir;
        this.parents = parents;
        if (!task.steps || task.steps.length === 0) {
            this.logDebug(chalk_1.default.gray('No actions have been specified for this task.'));
            return;
        }
        this.env = this.resolveEnvironment(parents);
        const envlogs = [];
        for (const [k, v] of Object.entries(this.env)) {
            const vv = v !== null && v !== void 0 ? v : '';
            const trimmed = vv.length > ENV_TRIM_LEN ? vv.substr(0, ENV_TRIM_LEN) + '...' : vv;
            envlogs.push(`${k}=${trimmed}`);
        }
        if (envlogs.length) {
            this.logDebug(chalk_1.default.gray(`${chalk_1.default.underline('env')}: ${envlogs.join(' ')}`));
        }
        // evaluate condition
        if (!this.evalCondition(task)) {
            this.log('condition exited with non-zero - skipping');
            return;
        }
        // verify we required environment variables are defined
        const merged = Object.assign(Object.assign({}, process.env), this.env);
        const missing = [];
        for (const name of (_b = task.requiredEnv) !== null && _b !== void 0 ? _b : []) {
            if (!(name in merged)) {
                missing.push(name);
            }
        }
        if (missing.length > 0) {
            throw new Error(`missing required environment variables: ${missing.join(',')}`);
        }
        this.execute(task);
    }
    /**
     * Executes steps in a given task
     */
    execute(task) {
        for (const step of task.steps) {
            if (step.say) {
                (0, logging_1.info)(this.fmtLog(step.say));
            }
            if (step.spawn) {
                this.runtime.runTask(step.spawn, [...this.parents, this.task.name]);
            }
            const execs = step.exec ? [{ command: step.exec, label: step.exec }] : [];
            if (step.builtin) {
                const builtin = this.resolveBuiltin(step.builtin);
                execs.push(builtin);
            }
            for (const exec of execs) {
                let hasError = false;
                const cwd = step.cwd;
                try {
                    const result = this.shell({
                        command: exec.command,
                        label: exec.label,
                        cwd,
                    });
                    hasError = result.status !== 0;
                }
                catch (e) {
                    throw e;
                }
                if (hasError) {
                    throw new Error(`Task "${this.fullname}" failed when executing command "${exec.command}" (cwd: ${(0, path_1.resolve)(cwd !== null && cwd !== void 0 ? cwd : this.workdir)})`);
                }
            }
        }
    }
    /**
     * Determines if a task should be executed based on "condition".
     *
     * @returns true if the task should be executed or false if the condition
     * evaluates to false (exits with non-zero), indicating that the task should
     * be skipped.
     */
    evalCondition(task) {
        // no condition, carry on
        if (!task.condition) {
            return true;
        }
        this.log(chalk_1.default.gray(`${chalk_1.default.underline('condition')}: ${task.condition}`));
        const result = this.shell({
            command: task.condition,
            logprefix: 'condition: ',
            quiet: true,
        });
        if (result.status === 0) {
            return true;
        }
        else {
            return false;
        }
    }
    /**
     * Renders the runtime environment for a task. Namely, it supports this syntax
     * `$(xx)` for allowing environment to be evaluated by executing a shell
     * command and obtaining its result.
     */
    resolveEnvironment(parents) {
        var _a, _b, _c, _d, _e, _f;
        let env = (_a = this.runtime.manifest.env) !== null && _a !== void 0 ? _a : {};
        // add env from all parent tasks one by one
        for (const parent of parents) {
            env = Object.assign(Object.assign({}, env), ((_c = (_b = this.runtime.tryFindTask(parent)) === null || _b === void 0 ? void 0 : _b.env) !== null && _c !== void 0 ? _c : {}));
        }
        // apply the task's environment last
        env = Object.assign(Object.assign({}, env), ((_d = this.task.env) !== null && _d !== void 0 ? _d : {}));
        const output = {};
        for (const [key, value] of Object.entries(env !== null && env !== void 0 ? env : {})) {
            if (value.startsWith('$(') && value.endsWith(')')) {
                const query = value.substring(2, value.length - 1);
                const result = this.shellEval({ command: query });
                if (result.status !== 0) {
                    const error = result.error
                        ? result.error.stack
                        : (_f = (_e = result.stderr) === null || _e === void 0 ? void 0 : _e.toString()) !== null && _f !== void 0 ? _f : 'unknown error';
                    throw new Error(`unable to evaluate environment variable ${key}=${value}: ${error}`);
                }
                output[key] = result.stdout.toString('utf-8').trim();
            }
            else {
                output[key] = value;
            }
        }
        return output;
    }
    /**
     * Returns the "full name" of the task which includes all it's parent task names concatenated by chevrons.
     */
    get fullname() {
        return [...this.parents, this.task.name].join(' » ');
    }
    log(...args) {
        (0, logging_1.verbose)(this.fmtLog(...args));
    }
    logDebug(...args) {
        (0, logging_1.debug)(this.fmtLog(...args));
    }
    fmtLog(...args) {
        return (0, util_1.format)(`${chalk_1.default.underline(this.fullname)} |`, ...args);
    }
    shell(options) {
        var _a, _b;
        const quiet = (_a = options.quiet) !== null && _a !== void 0 ? _a : false;
        if (!quiet) {
            const log = [];
            if (options.logprefix) {
                log.push(options.logprefix);
            }
            log.push(options.label || options.command);
            if (options.cwd) {
                log.push(`(cwd: ${options.cwd})`);
            }
            this.log(log.join(' '));
        }
        const cwd = (_b = options.cwd) !== null && _b !== void 0 ? _b : this.workdir;
        if (!(0, fs_1.existsSync)(cwd) || !(0, fs_1.statSync)(cwd).isDirectory()) {
            throw new Error(`invalid workdir (cwd): ${cwd} must be an existing directory`);
        }
        return (0, child_process_1.spawnSync)(options.command, Object.assign(Object.assign(Object.assign({}, options), { cwd, shell: true, stdio: 'inherit', env: Object.assign(Object.assign({}, process.env), this.env) }), options.spawnOptions));
    }
    shellEval(options) {
        return this.shell(Object.assign(Object.assign({ quiet: true }, options), { spawnOptions: {
                stdio: ['inherit', 'pipe', 'inherit'],
            } }));
    }
    resolveBuiltin(builtin) {
        const task = `${builtin}.task.js`;
        const program = require.resolve((0, path_1.join)(__dirname, '../tasks', task));
        return {
            command: `${process.execPath} ${program}`,
            label: `${process.title} ${task}`,
        };
    }
}
exports.Task = Task;
